@@ -158,11 +158,145 @@ gtk_clutter_embed_send_configure (GtkClutterEmbed *embed)
   event->configure.y = allocation.y;
   event->configure.width = allocation.width;
   event->configure.height = allocation.height;
-  
+
   gtk_widget_event (widget, event);
   gdk_event_free (event);
 }
 
+#if defined(GDK_WINDOWING_WAYLAND) && defined(CLUTTER_WINDOWING_WAYLAND)
+static void
+gtk_clutter_embed_ensure_surface (GtkClutterEmbed *embed)
+{
+  GtkClutterEmbedPrivate *priv = embed->priv;
+
+  if (priv->subcompositor && !priv->clutter_surface)
+    {
+      GdkDisplay *display;
+      struct wl_compositor *compositor;
+
+      display = gtk_widget_get_display (GTK_WIDGET (embed));
+      compositor = gdk_wayland_display_get_wl_compositor (display);
+      priv->clutter_surface = wl_compositor_create_surface (compositor);
+    }
+}
+
+static void
+gtk_clutter_embed_ensure_subsurface (GtkClutterEmbed *embed)
+{
+  GtkClutterEmbedPrivate *priv;
+  GtkWidget *widget;
+  struct wl_surface *gtk_surface;
+  GdkWindow *window;
+  gint x, y;
+
+  widget = GTK_WIDGET (embed);
+  priv = embed->priv;
+
+  if (priv->subsurface)
+    return;
+
+  window = gtk_widget_get_window (widget);
+  gtk_surface = gdk_wayland_window_get_wl_surface (gdk_window_get_toplevel (window));
+  priv->subsurface =
+    wl_subcompositor_get_subsurface (priv->subcompositor,
+                                     priv->clutter_surface,
+                                     gtk_surface);
+
+  gdk_window_get_origin (window, &x, &y);
+  wl_subsurface_set_position (priv->subsurface, x, y);
+  wl_subsurface_set_desync (priv->subsurface);
+}
+#endif
+
+static void
+gtk_clutter_embed_ensure_stage_realized (GtkClutterEmbed *embed)
+{
+  GtkClutterEmbedPrivate *priv = GTK_CLUTTER_EMBED (embed)->priv;
+
+  if (!gtk_widget_get_realized (GTK_WIDGET (embed)))
+    return;
+
+  if (!clutter_actor_is_realized (priv->stage))
+    {
+      GdkWindow *window = gtk_widget_get_window (GTK_WIDGET (embed));
+
+#if defined(CLUTTER_WINDOWING_GDK)
+      if (clutter_check_windowing_backend (CLUTTER_WINDOWING_GDK))
+        {
+          clutter_gdk_set_stage_foreign (CLUTTER_STAGE (priv->stage), window);
+        }
+      else
+#endif
+#if defined(GDK_WINDOWING_X11) && defined(CLUTTER_WINDOWING_X11)
+      if (clutter_check_windowing_backend (CLUTTER_WINDOWING_X11) &&
+          GDK_IS_X11_WINDOW (window))
+        {
+          clutter_x11_set_stage_foreign (CLUTTER_STAGE (priv->stage),
+                                         GDK_WINDOW_XID (window));
+        }
+      else
+#endif
+#if defined(GDK_WINDOWING_WIN32) && defined(CLUTTER_WINDOWING_WIN32)
+      if (clutter_check_windowing_backend (CLUTTER_WINDOWING_WIN32) &&
+          GDK_IS_WIN32_WINDOW (window))
+        {
+          clutter_win32_set_stage_foreign (CLUTTER_STAGE (priv->stage),
+                                           GDK_WINDOW_HWND (window));
+        }
+      else
+#endif
+#if defined(GDK_WINDOWING_WAYLAND) && defined (CLUTTER_WINDOWING_WAYLAND)
+      if (clutter_check_windowing_backend (CLUTTER_WINDOWING_WAYLAND) &&
+          GDK_IS_WAYLAND_WINDOW (window))
+        {
+          gtk_clutter_embed_ensure_surface (embed);
+          clutter_wayland_stage_set_wl_surface (CLUTTER_STAGE (priv->stage),
+                                                priv->clutter_surface);
+        }
+      else
+#endif
+        {
+          g_warning ("No backend found!");
+        }
+
+      clutter_actor_realize (priv->stage);
+    }
+
+  /* A stage cannot really be unmapped because it is the top of
+   * Clutter's scene tree. So if the Gtk embedder is mapped, we
+   * translate this as visible for the ClutterStage. */
+  if (gtk_widget_get_mapped (GTK_WIDGET (embed)))
+    clutter_actor_show (priv->stage);
+
+  clutter_actor_queue_relayout (priv->stage);
+
+  gtk_clutter_embed_send_configure (embed);
+
+#if defined(GDK_WINDOWING_WAYLAND) && defined (CLUTTER_WINDOWING_WAYLAND)
+  if (clutter_check_windowing_backend (CLUTTER_WINDOWING_WAYLAND))
+    gtk_clutter_embed_ensure_subsurface (embed);
+#endif
+}
+
+static void
+gtk_clutter_embed_stage_unrealize (GtkClutterEmbed *embed)
+{
+  GtkClutterEmbedPrivate *priv = embed->priv;
+
+#if defined(GDK_WINDOWING_WAYLAND) && defined(CLUTTER_WINDOWING_WAYLAND)
+  g_clear_pointer (&priv->subsurface, wl_subsurface_destroy);
+  g_clear_pointer (&priv->clutter_surface, wl_surface_destroy);
+#endif
+
+  /* gtk may emit an unmap signal after dispose, so it's possible we
+   * may have already disposed priv->stage. */
+  if (priv->stage != NULL)
+    {
+      clutter_actor_hide (priv->stage);
+      clutter_actor_unrealize (priv->stage);
+    }
+}
+
 static void
 on_stage_queue_redraw (ClutterStage *stage,
                        ClutterActor *origin,
@@ -219,25 +353,9 @@ gtk_clutter_embed_dispose (GObject *gobject)
 static void
 gtk_clutter_embed_show (GtkWidget *widget)
 {
-  GtkClutterEmbedPrivate *priv = GTK_CLUTTER_EMBED (widget)->priv;
-
-  if (gtk_widget_get_realized (widget) && priv->stage != NULL)
-    clutter_actor_show (priv->stage);
-
   GTK_WIDGET_CLASS (gtk_clutter_embed_parent_class)->show (widget);
-}
 
-static void
-gtk_clutter_embed_hide (GtkWidget *widget)
-{
-  GtkClutterEmbedPrivate *priv = GTK_CLUTTER_EMBED (widget)->priv;
-
-  /* gtk emits a hide signal during dispose, so it's possible we may
-   * have already disposed priv->stage. */
-  if (priv->stage != NULL)
-    clutter_actor_hide (priv->stage);
-
-  GTK_WIDGET_CLASS (gtk_clutter_embed_parent_class)->hide (widget);
+  gtk_clutter_embed_ensure_stage_realized (GTK_CLUTTER_EMBED (widget));
 }
 
 static GdkWindow *
@@ -309,43 +427,9 @@ gtk_clutter_embed_draw (GtkWidget *widget, cairo_t *cr)
   return GTK_WIDGET_CLASS (gtk_clutter_embed_parent_class)->draw (widget, cr);
 }
 
-#if defined(GDK_WINDOWING_WAYLAND) && defined(CLUTTER_WINDOWING_WAYLAND)
-static void
-gtk_clutter_embed_ensure_subsurface (GtkClutterEmbed *embed)
-{
-  GtkClutterEmbedPrivate *priv;
-  GtkWidget *widget;
-  struct wl_surface *gtk_surface;
-  GtkAllocation allocation;
-  GdkWindow *window;
-  gint x, y;
-
-  widget = GTK_WIDGET (embed);
-  priv = embed->priv;
-
-  if (priv->subsurface)
-    return;
-
-  gtk_widget_get_allocation (widget, &allocation);
-  window = gtk_widget_get_window (widget);
-  gtk_surface = gdk_wayland_window_get_wl_surface (gdk_window_get_toplevel(window));
-
-  priv->subsurface =
-    wl_subcompositor_get_subsurface (priv->subcompositor,
-                                     priv->clutter_surface,
-                                     gtk_surface);
-
-  gtk_widget_translate_coordinates (widget, gtk_widget_get_toplevel (widget), 0, 0, &x, &y);
-  gdk_window_get_origin (gtk_widget_get_parent_window (widget), &x, &y);
-  wl_subsurface_set_position (priv->subsurface, x + allocation.x, y + allocation.y);
-  wl_subsurface_set_desync (priv->subsurface);
-}
-#endif
-
 static void
 gtk_clutter_embed_realize (GtkWidget *widget)
 {
-  GtkClutterEmbedPrivate *priv = GTK_CLUTTER_EMBED (widget)->priv;
   GtkAllocation allocation;
   GtkStyleContext *style_context;
   GdkWindow *window;
@@ -396,7 +480,7 @@ gtk_clutter_embed_realize (GtkWidget *widget)
   attributes.visual = gtk_widget_get_visual (widget);
 
   /* NOTE: GDK_MOTION_NOTIFY above should be safe as Clutter does its own
-   *       throttling. 
+   *       throttling.
    */
   attributes.event_mask = gtk_widget_get_events (widget)
                         | GDK_EXPOSURE_MASK
@@ -409,7 +493,8 @@ gtk_clutter_embed_realize (GtkWidget *widget)
                         | GDK_ENTER_NOTIFY_MASK
                         | GDK_LEAVE_NOTIFY_MASK
                         | GDK_TOUCH_MASK
-                        | GDK_SMOOTH_SCROLL_MASK;
+                        | GDK_SMOOTH_SCROLL_MASK
+                        | GDK_STRUCTURE_MASK;
 
   attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL;
 
@@ -430,19 +515,10 @@ gtk_clutter_embed_realize (GtkWidget *widget)
   style_context = gtk_widget_get_style_context (widget);
   gtk_style_context_set_background (style_context, window);
 
-#if defined(CLUTTER_WINDOWING_GDK)
-  if (clutter_check_windowing_backend (CLUTTER_WINDOWING_GDK))
-    {
-      clutter_gdk_set_stage_foreign (CLUTTER_STAGE (priv->stage), window);
-    }
-  else
-#endif
 #if defined(GDK_WINDOWING_X11) && defined(CLUTTER_WINDOWING_X11)
   if (clutter_check_windowing_backend (CLUTTER_WINDOWING_X11) &&
       GDK_IS_X11_WINDOW (window))
     {
-      clutter_x11_set_stage_foreign (CLUTTER_STAGE (priv->stage), GDK_WINDOW_XID (window));
-
       if (num_filter == 0)
         gdk_window_add_filter (NULL, gtk_clutter_filter_func, widget);
       num_filter++;
@@ -453,40 +529,23 @@ gtk_clutter_embed_realize (GtkWidget *widget)
   if (clutter_check_windowing_backend (CLUTTER_WINDOWING_WIN32) &&
       GDK_IS_WIN32_WINDOW (window))
     {
-      clutter_win32_set_stage_foreign (CLUTTER_STAGE (priv->stage), GDK_WINDOW_HWND (window));
-
       if (num_filter == 0)
         gdk_window_add_filter (NULL, gtk_clutter_filter_func, widget);
       num_filter++;
     }
+  else
 #endif
-
-#if defined(GDK_WINDOWING_WAYLAND) && defined(CLUTTER_WINDOWING_WAYLAND)
-  if (priv->subcompositor)
     {
-      GdkDisplay *display;
-      struct wl_compositor *compositor;
-
-      display = gtk_widget_get_display (widget);
-      compositor = gdk_wayland_display_get_wl_compositor (display);
-      priv->clutter_surface = wl_compositor_create_surface (compositor);
-      clutter_wayland_stage_set_wl_surface (CLUTTER_STAGE (priv->stage),
-                                            priv->clutter_surface);
+      /* Nothing to do. */
     }
-#endif
 
-  clutter_actor_realize (priv->stage);
-
-  if (gtk_widget_get_visible (widget))
-    clutter_actor_show (priv->stage);
-
-  gtk_clutter_embed_send_configure (GTK_CLUTTER_EMBED (widget));
+  gtk_clutter_embed_ensure_stage_realized (GTK_CLUTTER_EMBED (widget));
 }
 
 static void
 gtk_clutter_embed_unrealize (GtkWidget *widget)
 {
-  GtkClutterEmbedPrivate *priv = GTK_CLUTTER_EMBED (widget)->priv;
+  GtkClutterEmbed *embed = GTK_CLUTTER_EMBED (widget);
 
   if (num_filter > 0)
     {
@@ -495,8 +554,7 @@ gtk_clutter_embed_unrealize (GtkWidget *widget)
         gdk_window_remove_filter (NULL, gtk_clutter_filter_func, widget);
     }
 
-  if (priv->stage != NULL)
-    clutter_actor_hide (priv->stage);
+  gtk_clutter_embed_stage_unrealize (embed);
 
   GTK_WIDGET_CLASS (gtk_clutter_embed_parent_class)->unrealize (widget);
 }
@@ -520,6 +578,9 @@ gtk_clutter_embed_get_request_mode (GtkWidget *widget)
 	case CLUTTER_REQUEST_WIDTH_FOR_HEIGHT:
 	  mode = GTK_SIZE_REQUEST_WIDTH_FOR_HEIGHT;
 	  break;
+        case CLUTTER_REQUEST_CONTENT_SIZE:
+          mode = GTK_SIZE_REQUEST_CONSTANT_SIZE;
+          break;
 	}
     }
 
@@ -661,7 +722,8 @@ static gboolean
 gtk_clutter_embed_map_event (GtkWidget	 *widget,
                              GdkEventAny *event)
 {
-  GtkClutterEmbedPrivate *priv = GTK_CLUTTER_EMBED (widget)->priv;
+  GtkClutterEmbed *embed = GTK_CLUTTER_EMBED (widget);
+  GtkClutterEmbedPrivate *priv = embed->priv;
   GtkWidgetClass *parent_class;
   gboolean res = FALSE;
 
@@ -669,7 +731,7 @@ gtk_clutter_embed_map_event (GtkWidget	 *widget,
   if (parent_class->map_event)
     res = parent_class->map_event (widget, event);
 
-  clutter_actor_map (priv->stage);
+  gtk_clutter_embed_ensure_stage_realized (embed);
 
   clutter_actor_queue_redraw (priv->stage);
 
@@ -680,7 +742,7 @@ static gboolean
 gtk_clutter_embed_unmap_event (GtkWidget   *widget,
                                GdkEventAny *event)
 {
-  GtkClutterEmbedPrivate *priv = GTK_CLUTTER_EMBED (widget)->priv;
+  GtkClutterEmbed *embed = GTK_CLUTTER_EMBED (widget);
   GtkWidgetClass *parent_class;
   gboolean res = FALSE;
 
@@ -688,54 +750,11 @@ gtk_clutter_embed_unmap_event (GtkWidget   *widget,
   if (parent_class->unmap_event)
     res = parent_class->unmap_event (widget, event);
 
-  clutter_actor_unmap (priv->stage);
-
-#if defined(GDK_WINDOWING_WAYLAND) && defined(CLUTTER_WINDOWING_WAYLAND)
-  g_clear_pointer (&priv->subsurface, wl_subsurface_destroy);
-#endif
+  gtk_clutter_embed_stage_unrealize (embed);
 
   return res;
 }
 
-static void
-gtk_clutter_embed_map (GtkWidget *widget)
-{
-  GtkClutterEmbedPrivate *priv = GTK_CLUTTER_EMBED (widget)->priv;
-
-#if defined(GDK_WINDOWING_WAYLAND) && defined(CLUTTER_WINDOWING_WAYLAND)
-  {
-    GdkDisplay *gdk_display = gtk_widget_get_display (widget);
-
-    if (clutter_check_windowing_backend (CLUTTER_WINDOWING_WAYLAND) &&
-        GDK_IS_WAYLAND_DISPLAY (gdk_display))
-      {
-        gtk_clutter_embed_ensure_subsurface (GTK_CLUTTER_EMBED (widget));
-      }
-  }
-#endif
-
-  GTK_WIDGET_CLASS (gtk_clutter_embed_parent_class)->map (widget);
-
-  clutter_actor_map (priv->stage);
-}
-
-static void
-gtk_clutter_embed_unmap (GtkWidget *widget)
-{
-  GtkClutterEmbedPrivate *priv = GTK_CLUTTER_EMBED (widget)->priv;
-
-  GTK_WIDGET_CLASS (gtk_clutter_embed_parent_class)->unmap (widget);
-
-  /* gtk may emit an unmap signal after dispose, so it's possible we may
-   * have already disposed priv->stage. */
-  if (priv->stage != NULL)
-    clutter_actor_unmap (priv->stage);
-
-#if defined(GDK_WINDOWING_WAYLAND) && defined(CLUTTER_WINDOWING_WAYLAND)
-  g_clear_pointer (&priv->subsurface, wl_subsurface_destroy);
-#endif
-}
-
 static gboolean
 gtk_clutter_embed_focus_in (GtkWidget     *widget,
                             GdkEventFocus *event)
@@ -922,7 +941,7 @@ _gtk_clutter_embed_set_child_active (GtkClutterEmbed *embed,
       gdk_offscreen_window_set_embedder (child_window,
 					 NULL);
     }
-      
+
 }
 
 static void
@@ -1052,9 +1071,6 @@ gtk_clutter_embed_class_init (GtkClutterEmbedClass *klass)
   widget_class->realize = gtk_clutter_embed_realize;
   widget_class->unrealize = gtk_clutter_embed_unrealize;
   widget_class->show = gtk_clutter_embed_show;
-  widget_class->hide = gtk_clutter_embed_hide;
-  widget_class->map = gtk_clutter_embed_map;
-  widget_class->unmap = gtk_clutter_embed_unmap;
   widget_class->map_event = gtk_clutter_embed_map_event;
   widget_class->unmap_event = gtk_clutter_embed_unmap_event;
   widget_class->focus_in_event = gtk_clutter_embed_focus_in;
@@ -1137,11 +1153,6 @@ gtk_clutter_embed_init (GtkClutterEmbed *embed)
   /* we accept key focus */
   gtk_widget_set_can_focus (widget, TRUE);
 
-  G_GNUC_BEGIN_IGNORE_DEPRECATIONS
-  /* disable double-buffering: it's automatically provided by OpenGL */
-  gtk_widget_set_double_buffered (widget, FALSE);
-  G_GNUC_END_IGNORE_DEPRECATIONS
-
   /* we own the whole drawing of this widget, including the background */
   gtk_widget_set_app_paintable (widget, TRUE);
 
@@ -1252,7 +1263,7 @@ gtk_clutter_embed_set_use_layout_size (GtkClutterEmbed *embed,
   GtkClutterEmbedPrivate *priv = embed->priv;
 
   g_return_if_fail (GTK_CLUTTER_IS_EMBED (embed));
-  
+
   use_layout_size = !!use_layout_size;
   if (use_layout_size != priv->use_layout_size)
     {
@@ -1279,6 +1290,6 @@ gtk_clutter_embed_get_honor_stage_size (GtkClutterEmbed *embed)
   GtkClutterEmbedPrivate *priv = embed->priv;
 
   g_return_val_if_fail (GTK_CLUTTER_IS_EMBED (embed), FALSE);
-  
+
   return priv->use_layout_size;
 }
